"""
Utility for running re within disassembled code.

This utility extends and overwrites the existing re API to work correctly within the
context of a disassembler.
This module works just like the builtin re module, but adjusts offsets to be virtual addresses
and allows for searching specific segments.

usage::
    import dragodis
    from rugosa import re


    ptn = re.compile('some pattern')

    with dragodis.open_program("input.exe") as dis:
        for match in ptn.finditer(dis, '.text'):
            print('found marker at 0x{:0x}'.format(match.start()))
"""

import re
from typing import Iterable, List, Optional

import dragodis
from dragodis.interface import Function


class Match(object):
    """
    Wraps the SRE_Match object returned by re.
    """

    def __init__(self, match, seg_start):
        self._match = match
        self._start = seg_start

    def __getattr__(self, item):
        """
        Redirects anything that this class doesn't support back to the matchobject class

        :param item:

        :return:
        """
        return getattr(self._match, item, None)

    def start(self, group=None):
        """
        Returns the match object start value with respect to the segment start.

        :param group: optional group to obtain the start of

        :return: virtual start address
        """
        if group:
            _group_start = self._match.start(group)
            if _group_start < 0:  # group exists, but not contributing to match
                return _group_start

            return _group_start + self._start

        return self._match.start() + self._start

    def end(self, group=None):
        """
        Returns the match object end value with respect to the segment start.

        :param group: optional group to obtain the end of

        :return: virtual end address
        """
        if group:
            _group_end = self._match.end(group)
            if _group_end < 0:  # group exists, but not contributing to match
                return _group_end

            return _group_end + self._start

        return self._match.end() + self._start


class Pattern(object):
    """
    Wraps the SRE_Pattern object returned by re.
    """

    def __init__(self, ptn, flags=0):
        if isinstance(ptn, (str, bytes)):
            self._re = re.compile(ptn, flags=flags)
        else:
            self._re = ptn

    @property
    def pattern(self):
        return self._re.pattern

    def search(self, dis: dragodis.Disassembler, segment_name: str = None) -> Optional[Match]:
        """
        Performs the search functionality on the entire file, searching each segment individually.

        :param dis: Dragodis disassembler.
        :param segment_name: Restrict searching to segment with provided name
        :return: match object modified to match the segment start address
        """
        segments = [dis.get_segment(segment_name)] if segment_name else dis.segments
        for segment in segments:
            match = self._re.search(segment.data)
            if match:
                return Match(match, segment.start)
        return None

    def finditer(self, dis: dragodis.Disassembler, segment_name=None) -> Iterable[Match]:
        """
        Performs the finditer functionality on the entire file, searching each segment individually.

        :param dis: Dragodis disassembler.
        :param segment_name: Restrict searching to segment with provided name

        :yield: match object
        """
        segments = [dis.get_segment(segment_name)] if segment_name else dis.segments
        for segment in segments:
            for match in self._re.finditer(segment.data):
                yield Match(match, segment.start)

    def findall(self, dis: dragodis.Disassembler, segment_name=None) -> List[Match]:
        """
        Performs the findall functionality on the entire file.

        :param dis: Dragodis disassembler.
        :param segment_name: Restrict searching to segment with provided name
        :return: list of match objects
        """
        segments = [dis.get_segment(segment_name)] if segment_name else dis.segments
        matches = []
        for segment in segments:
            matches.extend(self._re.findall(segment.data))
        return matches

    def find_functions(self, dis: dragodis.Disassembler, segment_name: str = None) -> Iterable[Function]:
        """
        Uses finditer() to search for functions that contains a match for the given
        regular expression pattern.

        :param dis: Dragodis disassembler.
        :param segment_name: Restrict searching to segment with provided name
        :yields: Function object for each function.
        """
        cache = set()
        for match in self.finditer(dis, segment_name):
            try:
                func = dis.get_function(match.start())
            except dragodis.NotExistError:
                continue

            if func.start not in cache:
                cache.add(func.start)
                yield func


def compile(pattern, flags=0):
    """Compile a regular expression returning a Pattern object."""
    return Pattern(pattern, flags=flags)


def search(pattern, dis: dragodis.Disassembler, segment_name: str = None, flags=0):
    """Search with regular expression, returning a Match object."""
    return Pattern(pattern, flags=flags).search(dis, segment_name=segment_name)


def finditer(pattern, dis: dragodis.Disassembler, segment_name: str = None, flags=0):
    """Iterator of non-overlapping matches."""
    ptn = Pattern(pattern, flags=flags)
    yield from ptn.finditer(dis, segment_name=segment_name)


def findall(pattern, dis: dragodis.Disassembler, segment_name: str = None, flags=0):
    """Returns a list of non-overlapping matches."""
    return Pattern(pattern, flags=flags).findall(dis, segment_name=segment_name)


def find_functions(pattern, dis: dragodis.Disassembler, segment_name: str = None, flags=0) -> Iterable[Function]:
    """
    Uses finditer() to search for functions that contains a match for the given
    regular expression pattern.

    Yields dragodis.interface.Function object for each function.
    """
    if isinstance(pattern, Pattern):
        _regex = pattern
    else:
        _regex = Pattern(pattern, flags)

    yield from _regex.find_functions(dis, segment_name=segment_name)



"""
Interface for operand management in x86.
"""

import logging
from typing import Optional

from dragodis import OperandType
from dragodis.interface import Phrase, MemoryReference, Immediate, Register

from .. import utils
from ..operands import Operand

logger = logging.getLogger(__name__)


class x86_64Operand(Operand):

    def _calc_displacement(self):
        """
        Calculate the displacement offset of the operand's text.

        e.g:
            word ptr [rdi+rbx]

        :return int: calculated value
        """
        addr = self.base + self.index * self.scale + self.offset
        if addr < 0:
            addr = utils.unsigned(addr, self._cpu_context.byteness)
        logger.debug(
            "Calculating operand: %s -> 0x%X + 0x%X*0x%X %s 0x%X = 0x%X" % (
                self.text,
                self.base,
                self.index,
                self.scale,
                "-" if self.offset < 0 else "+",
                abs(self.offset),
                addr
            )
        )

        return addr

    @property
    def scale(self) -> Optional[int]:
        """
        The scaling factor of the index if operand is a displacement.

        e.g.
            [ebp+ecx*2+var_8] -> 2
        """
        phrase = self._operand.value
        if not isinstance(phrase, Phrase):
            return None
        return phrase.scale

    @property
    def index(self) -> Optional[int]:
        """
        The value of the index register if operand is a displacement.
        Returns None if not a displacement.

        e.g.
            [ebp+ecx*2+var_8] -> ecx
        """
        phrase = self._operand.value
        if not isinstance(phrase, Phrase):
            return None
        index_reg = phrase.index
        if not index_reg:
            return 0
        value = self._cpu_context.registers[index_reg.name]
        return utils.signed(value, self._cpu_context.byteness)

    @property
    def offset(self) -> Optional[int]:
        """
        The offset value if the operand is a displacement.

        e.g.
            [ebp+ecx*2+8] -> 8
            fs:[eax] -> eax
        """
        phrase = self._operand.value
        if not isinstance(phrase, Phrase):
            return None
        offset = phrase.offset
        if isinstance(offset, Register):
            offset = self._cpu_context.registers[offset.name]
        return offset

    @property
    def addr(self) -> Optional[int]:
        """
        Retrieves the referenced memory address of the operand.

        :return int: Memory address or None if operand is not a memory reference.
        """
        addr = None
        value = self._operand.value
        if isinstance(value, Phrase):
            # These need to be handled in the same way even if they don't contain the same types of data...
            addr = self._calc_displacement()

        # TODO: Originally this only pulled o_mem types, but now we also do o_near/o_far
        #   Determine if that is okay.
        elif isinstance(value, (MemoryReference, Immediate)):
            addr = int(value)

        if addr is not None:
            logger.debug("&%s -> 0x%X", self.text, addr)

        return addr

    @property
    def base_addr(self):
        """
        Retrieves the referenced memory address of the operand minus any indexing that
        has occurred.

        This is useful for pulling out the un-offseted address within a loop.
        e.g. "movzx   edx, [ebp+ecx*2+var_8]"
        where ecx is the loop index starting at a non-zero value.

        :return int: Memory address or None if operand is not a memory reference.
        """
        addr = self.addr
        if addr is None:
            return None
        index = self.index
        if index is not None:
            addr -= index * self.scale
        return addr

    @property
    def value(self):
        return super().value

    @value.setter
    def value(self, value):
        try:
            logger.debug("0x%X -> %s", value, self.text)
        except TypeError:
            logger.debug("%r -> %s", value, self.text)

        # On 64-bit, the destination register must be set to 0 first (per documentation)
        # TODO: Check if this happens regardless of the source size
        if (
                self.type == OperandType.register
                and self._cpu_context.bitness == 64
                and self.width == 4
        ):  # Only do this for 32-bit setting
            reg_name = self._operand.value.name.lower()
            self._cpu_context.registers.clear_family(reg_name)

        super(x86_64Operand, self.__class__).value.__set__(self, value)



"""
ARM opcodes
"""

import logging

from .. import utils
from ..cpu_context import ProcessorContext
from ..instruction import Instruction
from ..exceptions import EmulationError
from ..registry import registrar
from . import utils as arm_utils


logger = logging.getLogger(__name__)


# Dictionary containing opcode names -> function
OPCODES = {}
opcode = registrar(OPCODES, name="opcode")


@opcode
def NOP(cpu_context: ProcessorContext, instruction: Instruction):
    return


#region Conditional branch instructions


@opcode
def CBNZ(cpu_context: ProcessorContext, instruction: Instruction):
    """Compare and branch if nonzero"""
    operands = instruction.operands
    value = operands[0].value
    jump_target = operands[1].value

    if value != 0:
        cpu_context.ip = jump_target

    # TODO: Update branch tracking.


@opcode
def CBZ(cpu_context: ProcessorContext, instruction: Instruction):
    """Compare and branch if zero"""
    operands = instruction.operands
    value = operands[0].value
    jump_target = operands[1].value

    if value == 0:
        cpu_context.ip = jump_target

    # TODO: Update branch tracking.


@opcode
def TBNZ(cpu_context: ProcessorContext, instruction: Instruction):
    """Test bit and branch if nonzero"""
    operands = instruction.operands
    value = operands[0].value
    bit_number = operands[1].value
    jump_target = operands[2].value

    if value & (1 << bit_number) != 0:
        cpu_context.ip = jump_target

    # TODO: Update branch tracking.


@opcode
def TBZ(cpu_context: ProcessorContext, instruction: Instruction):
    """Test bit and branch if zero"""
    operands = instruction.operands
    value = operands[0].value
    bit_number = operands[1].value
    jump_target = operands[2].value

    if value & (1 << bit_number) == 0:
        cpu_context.ip = jump_target

    # TODO: Update branch tracking.

#endregion

#region Unconditional branch (immediate/register)

@opcode("b")
@opcode("br")
@opcode("bx")
def B(cpu_context: ProcessorContext, instruction: Instruction):
    """Branch unconditionally"""
    cpu_context.ip = instruction.operands[0].value


@opcode("bl")
@opcode("blr")
@opcode("blx")
def BL(cpu_context: ProcessorContext, instruction: Instruction):
    """Branch with link"""
    operands = instruction.operands
    # Function pointer can be a memory reference or immediate.
    func_ea = operands[0].addr or operands[0].value
    # Using signature to get demangled name.
    func_name = cpu_context.emulator.disassembler.get_function_signature(func_ea).name

    logger.debug("call %s", func_name or f"0x{func_ea:X}")

    # Store next ip to lr register.
    cpu_context.registers.lr = instruction.ip + 4
    # Branch
    cpu_context.ip = operands[0].value

    if operands[0].is_func_ptr:
        cpu_context._execute_call(func_ea, func_name, instruction.ip)

    # Restore instruction pointer to return address.
    cpu_context.ip = cpu_context.registers.lr

#endregion

#region System register instructions


@opcode("mrs")
@opcode("msr")
def _mov(cpu_context: ProcessorContext, instruction: Instruction):
    operands = instruction.operands
    operands[0].value = operands[1].value

#endregion

#region Load/Store register

@opcode
def LDR(cpu_context: ProcessorContext, instruction: Instruction):
    """Load with immediate"""
    operands = instruction.operands
    operands[0].value = operands[1].value


@opcode
def STR(cpu_context: ProcessorContext, instruction: Instruction):
    """Store with immediate"""
    operands = instruction.operands
    operands[1].value = operands[0].value

#endregion

#region Load/Store register (unscaled offset)

# TODO: Can LDUR just be an alias for LDR?
@opcode
def LDUR(cpu_context: ProcessorContext, instruction: Instruction):
    """Load register (unscaled offset)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


# TODO: Can STUR just be an alias for STR?
@opcode
def STUR(cpu_context: ProcessorContext, instruction: Instruction):
    """Store register (unscaled offset)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Load Multiple (Increment After, Full Descending)


def _pop(cpu_context, reg_list):
    for reg_name in reg_list:
        data = cpu_context.memory.read(cpu_context.sp, 4)
        value = int.from_bytes(data, cpu_context.byteorder)
        cpu_context.registers[reg_name] = value
        cpu_context.sp += 4


@opcode
def LDM(cpu_context: ProcessorContext, instruction: Instruction):
    """Load Multiple Registers"""
    operands = instruction.operands

    # If we only have one operand, this is Ghidra's version of a POP.
    if len(operands) == 1 and operands[0].text.startswith("sp!"):
        _pop(cpu_context, operands[0].register_list[1:])
        return

    reg_list = operands[1].register_list
    if not reg_list:
        raise EmulationError(f"Expected {operands[1].text} to be a register list.")

    for reg_name in reg_list:
        data = cpu_context.memory.read(operands[0].value, 4)
        value = int.from_bytes(data, cpu_context.byteorder)
        cpu_context.registers[reg_name] = value
        # TODO: confirm operand is auto-increased in ARMInstruction._execute()


@opcode
def POP(cpu_context: ProcessorContext, instruction: Instruction):
    """Load Multiple Register from Stack"""
    operands = instruction.operands
    reg_list = operands[0].register_list
    if not reg_list:
        raise EmulationError(f"Expected {operands[0].text} to be a register list.")

    _pop(cpu_context, reg_list)


#endregion

#region Store Multiple (Increment After, Empty Ascending)


def _push(cpu_context, reg_values):
    # .value is a list of register values from lowest to highest.
    # Registers are pushed from largest to smallest for PUSH.
    for value in reversed(reg_values):
        cpu_context.sp -= 4
        cpu_context.memory.write_data(cpu_context.sp, value)


@opcode
def STM(cpu_context: ProcessorContext, instruction: Instruction):
    """Store Multiple Registers"""
    operands = instruction.operands

    # If we only have one operand, this is Ghidra's version of a PUSH.
    if len(operands) == 1 and operands[0].text.startswith("sp!"):
        _push(cpu_context, operands[0].value[1:])
        return

    reg_values = operands[1].value  # .value is a list of values
    if not isinstance(reg_values, list):
        raise EmulationError(f"Expected {operands[1].text} to be a register list.")

    for value in reversed(reg_values):
        cpu_context.memory.write_data(operands[0].value, value)
        # TODO: confirm operand is auto-decreased in ARMInstruction._execute()

@opcode
def PUSH(cpu_context: ProcessorContext, instruction: Instruction):
    """Store Multiple Register onto Stack"""
    operands = instruction.operands
    reg_values = operands[0].value
    if not isinstance(reg_values, list):
        raise EmulationError(f"Expected {operands[0].text} to be a register list.")

    _push(cpu_context, reg_values)

#endregion

#region Load/Store Pair


@opcode
def LDP(cpu_context: ProcessorContext, instruction: Instruction):
    """Load Pair"""
    operands = instruction.operands
    value_a = operands[2].value
    data = cpu_context.memory.read(
        operands[2].addr + operands[0].width,
        operands[1].width,
    )
    value_b = int.from_bytes(data, cpu_context.byteorder)

    logger.debug("Load 0x%X into %s", value_a, operands[0].text)
    logger.debug("Load 0x%X into %s", value_b, operands[1].text)

    operands[0].value = value_a
    operands[1].value = value_b


@opcode
def STP(cpu_context: ProcessorContext, instruction: Instruction):
    """Store Pair"""
    operands = instruction.operands
    value_a = operands[0].value
    value_b = operands[1].value

    logger.debug("Store 0x%X and 0x%X into %s", value_a, value_b, operands[2].text)

    operands[2].value = value_a
    cpu_context.memory.write(
        operands[2].addr + operands[0].width,
        value_b.to_bytes(operands[1].width, cpu_context.byteorder),
    )

#endregion

#region Load/Store unprivileged

@opcode
def LDTR(cpu_context: ProcessorContext, instruction: Instruction):
    """Load unprivileged register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STTR(cpu_context: ProcessorContext, instruction: Instruction):
    """Store unprivileged register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Load-Exclusive/Store-Exclusive


@opcode
def LDXR(cpu_context: ProcessorContext, instruction: Instruction):
    """Load Exclusive register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LDXP(cpu_context: ProcessorContext, instruction: Instruction):
    """Load Exclusive pair"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STXR(cpu_context: ProcessorContext, instruction: Instruction):
    """Store Exclusive register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STXP(cpu_context: ProcessorContext, instruction: Instruction):
    """Store Exclusive pair"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


#endregion

#region Load-Acquire/Store-Release


@opcode
def LDAPR(cpu_context: ProcessorContext, instruction: Instruction):
    """Load-Acquire RCpc Register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LDAPUR(cpu_context: ProcessorContext, instruction: Instruction):
    """Load-Acquire RCpc Register (unscaled)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LDAR(cpu_context: ProcessorContext, instruction: Instruction):
    """Load-Acquire Register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STLR(cpu_context: ProcessorContext, instruction: Instruction):
    """Store-Release Register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STLUR(cpu_context: ProcessorContext, instruction: Instruction):
    """Store-Release Register (unscaled)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LDAXR(cpu_context: ProcessorContext, instruction: Instruction):
    """Load-Acquire Exclusive register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LDAXP(cpu_context: ProcessorContext, instruction: Instruction):
    """Load-Acquire Exclusive pair"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STLXR(cpu_context: ProcessorContext, instruction: Instruction):
    """Store-Release Exclusive register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STLXP(cpu_context: ProcessorContext, instruction: Instruction):
    """Store-Release Exclusive pair"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


#endregion

#region The LoadLOAcquire/StoreLORelease


@opcode
def LDLAR(cpu_context: ProcessorContext, instruction: Instruction):
    """LoadLOAcquire register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STLLR(cpu_context: ProcessorContext, instruction: Instruction):
    """StoreLORelease register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region The Load/Store SIMD and Floating-point Non-temporal pair


@opcode
def LDNP(cpu_context: ProcessorContext, instruction: Instruction):
    """Load pair of scalar SIMD&FP registers"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STNP(cpu_context: ProcessorContext, instruction: Instruction):
    """Store pair of scalar SIMD&FP registers"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


#endregion

#region Load/Store Vector


@opcode
def LD1(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Load single 1-element structure to one lane of one register LD1 (single structure) on page C7-1637
    Load multiple 1-element structures to one register or to two, three, or four consecutive registers
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LD2(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Load single 2-element structure to one lane of two consecutive registers LD2 (single structure)
    Load multiple 2-element structures to two consecutive registers
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LD3(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Load single 3-element structure to one lane of three consecutive registers
    Load multiple 3-element structures to three consecutive registers
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LD4(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Load single 4-element structure to one lane of four consecutive registers
    Load multiple 4-element structures to four consecutive registers
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def ST1(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Store single 1-element structure from one lane of one register
    Store multiple 1-element structures from one register, or from two, three, or four consecutive registers
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def ST2(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Store single 2-element structure from one lane of two consecutive registers
    Store multiple 2-element structures from two consecutive registers
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def ST3(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Store single 3-element structure from one lane of three consecutive registers
    Store multiple 3-element structures from three consecutive registers
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def ST4(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Store single 4-element structure from one lane of four consecutive registers
    Store multiple 4-element structures from four consecutive registers
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LD1R(cpu_context: ProcessorContext, instruction: Instruction):
    """Load single 1-element structure and replicate to all lanes of one register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LD2R(cpu_context: ProcessorContext, instruction: Instruction):
    """Load single 2-element structure and replicate to all lanes of two registers"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LD3R(cpu_context: ProcessorContext, instruction: Instruction):
    """Load single 3-element structure and replicate to all lanes of three registers"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LD4R(cpu_context: ProcessorContext, instruction: Instruction):
    """Load single 4-element structure and replicate to all lanes of four registers"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Compare and Swap


@opcode
def CAS(cpu_context: ProcessorContext, instruction: Instruction):
    """Compare and swap"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CASP(cpu_context: ProcessorContext, instruction: Instruction):
    """Compare and swap pair"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


#endregion

#region Atomic memory operations


@opcode
def LDADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic add"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LDCLR(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic bit clear"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LDEOR(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic exclusive OR"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LDSET(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic bit set"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LDMAX(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic signed maximum"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LDMIN(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic signed minimum"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LDUMAX(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic unsigned maximum"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LDUMIN(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic unsigned minimum"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic add, without return"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STCLR(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic bit clear, without return"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STEOR(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic exclusive OR, without return"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STSET(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic bit set, without return"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STMAX(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic signed maximum, without return"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STMIN(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic signed minimum, without return"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STUMAX(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic unsigned maximum, without return"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def STUMIN(cpu_context: ProcessorContext, instruction: Instruction):
    """Atomic unsigned minimum, without return"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


#endregion

#region Swap


@opcode
def SWP(cpu_context: ProcessorContext, instruction: Instruction):
    """Swap"""
    operands = instruction.operands
    value_b = operands[1].value
    value_c = operands[2].value

    logger.debug("Swap %s %s %s", operands[0].text, operands[1].text, operands[2].text)

    operands[0].value = value_c
    operands[2].value = value_b


#endregion

#region Arithmetic (immediate)


@opcode("add")
@opcode("adc")
def ADD(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Handle both ADC and ADD here since the only difference is the flags.
    """
    operands = instruction.operands
    term_1 = operands[-2].value
    term_2 = operands[-1].value
    result = term_1 + term_2
    if instruction.root_mnem.startswith("adc"):
        result += cpu_context.registers.c

    width = get_max_operand_size(operands)
    mask = utils.get_mask(width)

    if instruction.flag_update:
        cpu_context.registers.c = int(result > mask)
        cpu_context.registers.z = int(result & mask == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.registers.v = int(utils.sign_bit(~(term_1 ^ term_2) & (term_2 ^ result), width) == 0)
        cpu_context.jcccontext.update_flag_opnds(["c", "z", "n", "v"], operands)

    result = result & mask

    logger.debug("0x%X + 0x%X = 0x%X", term_1, term_2, result)
    operands[0].value = result


# TODO: Due to simplification, it may be better to just keep the opcodes separate.
@opcode("sub")
@opcode("sbc")
@opcode("rsb")
@opcode("rsc")
def SUB(cpu_context: ProcessorContext, instruction: Instruction):
    """Subtract"""
    operands = instruction.operands
    term_1 = operands[1].value
    term_2 = operands[2].value
    if instruction.mnem.startswith("r"):  # reverse subtract
        term_1, term_2 = term_2, term_1

    result = term_1 - term_2
    if instruction.mnem.startswith(("sbc", "rsc")):
        result -= cpu_context.registers.c ^ 1

    if instruction.flag_update:
        width = get_max_operand_size(operands)
        mask = utils.get_mask(width)
        cpu_context.registers.c = int((term_1 & mask) < (term_2 & mask))
        cpu_context.registers.z = int(result & mask == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.registers.v = int(utils.sign_bit((term_1 ^ term_2) & (term_1 ^ result), width) == 0)
        cpu_context.jcccontext.update_flag_opnds(["c", "z", "n", "v"], operands)

    logger.debug("0x%X - 0x%X = 0x%X", term_1, term_2, result)
    operands[0].value = result


@opcode
def CMP(cpu_context: ProcessorContext, instruction: Instruction):
    """Compare"""
    operands = instruction.operands
    term_1 = operands[0].value
    term_2 = operands[1].value
    result = term_1 - term_2
    width = get_max_operand_size(operands)

    # Flags are always updated for CMP
    mask = utils.get_mask(width)
    cpu_context.registers.c = int((term_1 & mask) < (term_2 & mask))
    cpu_context.registers.z = int(result & mask == 0)
    cpu_context.registers.n = utils.sign_bit(result, width)
    cpu_context.registers.v = int(utils.sign_bit((term_1 ^ term_2) & (term_1 ^ result), width) == 0)
    cpu_context.jcccontext.update_flag_opnds(["c", "z", "n", "v"], operands)

    logger.debug("0x%X <-> 0x%X = 0x%X", term_1, term_2, result)


@opcode
def CMN(cpu_context: ProcessorContext, instruction: Instruction):
    """Compare negative"""
    operands = instruction.operands
    value_a = operands[1].value
    value_b = operands[2].value
    result = value_a + value_b
    width = get_max_operand_size(operands)

    mask = utils.get_mask(width)
    cpu_context.registers.c = int(result > mask)
    cpu_context.registers.z = int(result & mask == 0)
    cpu_context.registers.n = utils.sign_bit(result, width)
    cpu_context.registers.v = int(utils.sign_bit(~(value_a ^ value_b) & (value_b ^ result), width) == 0)
    cpu_context.jcccontext.update_flag_opnds(["c", "z", "n", "v"], operands)

    logger.debug("0x%X <-> 0x%X = 0x%X", value_a, value_b, result)


#endregion

#region Logical (immediate)


@opcode
def AND(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitwise AND"""
    operands = instruction.operands
    opvalue2 = operands[1].value
    opvalue3 = operands[2].value
    result = opvalue2 & opvalue3

    if instruction.flag_update:
        width = get_max_operand_size(operands)
        cpu_context.registers.z = int(result == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.jcccontext.update_flag_opnds(["z", "n"], operands)

    logger.debug("0x%X & 0x%X = 0x%X", opvalue2, opvalue3, result)
    operands[0].value = result


def TST(cpu_context: ProcessorContext, instruction: Instruction):
    """Test bits (same as ANDS, but result is discarded)"""
    operands = instruction.operands
    opvalue2 = operands[1].value
    opvalue3 = operands[2].value
    result = opvalue2 & opvalue3

    width = get_max_operand_size(operands)
    cpu_context.registers.z = int(result == 0)
    cpu_context.registers.n = utils.sign_bit(result, width)
    cpu_context.jcccontext.update_flag_opnds(["z", "n"], operands)

    logger.debug("0x%X & 0x%X = 0x%X", opvalue2, opvalue3, result)


@opcode
def EOR(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitwise exclusive OR"""
    operands = instruction.operands
    opvalue2 = operands[1].value
    opvalue3 = operands[2].value
    result = opvalue2 ^ opvalue3

    if instruction.flag_update:
        width = get_max_operand_size(operands)
        cpu_context.registers.z = int(result == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.jcccontext.update_flag_opnds(["z", "n"], operands)

    logger.debug("0x%X ^ 0x%X = 0x%X", opvalue2, opvalue3, result)
    operands[0].value = result


@opcode
def TEQ(cpu_context: ProcessorContext, instruction: Instruction):
    """Test Equivalence (same as EORS, except the result is discarded)"""
    operands = instruction.operands
    opvalue2 = operands[1].value
    opvalue3 = operands[2].value
    result = opvalue2 ^ opvalue3

    width = get_max_operand_size(operands)
    cpu_context.registers.z = int(result == 0)
    cpu_context.registers.n = utils.sign_bit(result, width)
    cpu_context.jcccontext.update_flag_opnds(["z", "n"], operands)

    logger.debug("0x%X ^ 0x%X = 0x%X", opvalue2, opvalue3, result)


@opcode
def ORR(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitwise inclusive OR"""
    operands = instruction.operands
    opvalue2 = operands[1].value
    opvalue3 = operands[2].value
    result = opvalue2 | opvalue3

    if instruction.flag_update:
        width = get_max_operand_size(operands)
        cpu_context.registers.z = int(result == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.jcccontext.update_flag_opnds(["z", "n"], operands)

    logger.debug("0x%X | 0x%X = 0x%X", opvalue2, opvalue3, result)
    operands[0].value = result

#endregion

#region Move (wide immediate)


@opcode("mov")
@opcode("movz")
@opcode("cpy")
def MOV(cpu_context: ProcessorContext, instruction: Instruction):
    """Move wide with zero"""
    operands = instruction.operands
    result = operands[1].value

    if instruction.flag_update:
        width = get_max_operand_size(operands)
        cpu_context.registers.z = int(result == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.jcccontext.update_flag_opnds(["z", "n"], operands)

    operands[0].value = result


@opcode
def MOVN(cpu_context: ProcessorContext, instruction: Instruction):
    """Move wide with NOT"""
    operands = instruction.operands
    result = ~operands[1].value

    if instruction.flag_update:
        width = get_max_operand_size(operands)
        cpu_context.registers.z = int(result == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.jcccontext.update_flag_opnds(["z", "n"], operands)

    operands[0].value = result


@opcode
def MOVK(cpu_context: ProcessorContext, instruction: Instruction):
    """Move wide with keep"""
    operands = instruction.operands

    # TODO: Is it always lsl?
    shift_mask = {
        0: 0xFFFFFFFFFFFF0000,
        16: 0xFFFFFFFF0000FFFF,
        32: 0xFFFF0000FFFFFFFF,
        48: 0x0000FFFFFFFFFFFF,
    }[operands[1].shift_count]
    operands[0].value = (operands[0].value & shift_mask) | operands[1].value


#endregion

#region PC-relative address calculation


@opcode
def ADRP(cpu_context: ProcessorContext, instruction: Instruction):
    """Compute address of 4KB page at a PC-relative offset"""
    operands = instruction.operands
    pc = cpu_context.registers.pc
    pc = pc & 0xFFFFFFFFFFFFF000  # Zero out bottom 12 bits of PC
    opvalue2 = operands[1].value
    result = pc + 0x1000*opvalue2

    logger.debug("0x%X + 0x1000*0x%X = 0x%X", pc, opvalue2, result)
    operands[0].value = result


@opcode
def ADR(cpu_context: ProcessorContext, instruction: Instruction):
    """Compute address of label at a PC-relative offset."""
    operands = instruction.operands
    pc = cpu_context.registers.pc
    opvalue2 = operands[1].value
    result = pc + opvalue2

    logger.debug("0x%X + 0x%X = 0x%X", pc, opvalue2, result)
    operands[0].value = result

#endregion

#region Bitfield move


@opcode
def BFM(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitfield move"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SBFM(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed bitfield move"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UBFM(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned bitfield move (32-bit)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Bitfield insert and extract


@opcode
def BFC(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitfield insert clear"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def BFI(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitfield insert"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def BFXIL(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitfield extract and insert low"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SBFIZ(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed bitfield insert in zero"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SBFX(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed bitfield extract"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UBFIZ(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned bitfield insert in zero"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UBFX(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned bitfield extract"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


#endregion

#region Extract register


@opcode
def EXTR(cpu_context: ProcessorContext, instruction: Instruction):
    """Extract register from pair"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


#endregion

#region Shift (immediate)


@opcode
def ASR(cpu_context: ProcessorContext, instruction: Instruction):
    """Arithmetic shift right"""
    operands = instruction.operands
    value = operands[1].value
    count = operands[2].value

    width = get_max_operand_size(operands)
    carry, result = arm_utils.asr(value, count, width=width)

    if instruction.flag_update:
        if count:  # C register is unaffected if the shift value is 0
            cpu_context.registers.c = carry
        cpu_context.registers.z = int(result == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.jcccontext.update_flag_opnds(["c", "z", "n"], operands)

    logger.debug("(0x%X >> 0x%X) = 0x%X", value, count, result)
    operands[0].value = result


@opcode
def LSL(cpu_context: ProcessorContext, instruction: Instruction):
    """Logical shift left"""
    operands = instruction.operands
    value = operands[1].value
    count = operands[2].value

    width = get_max_operand_size(operands)
    carry, result = arm_utils.lsl(value, count, width=width)

    if instruction.flag_update:
        if count:
            cpu_context.registers.c = carry
        cpu_context.registers.z = int(result == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.jcccontext.update_flag_opnds(["c", "z", "n"], operands)

    logger.debug("(0x%X << 0x%X) = 0x%X", value, count, result)
    operands[0].value = result


@opcode
def LSR(cpu_context: ProcessorContext, instruction: Instruction):
    """Logical shift right"""
    operands = instruction.operands
    value = operands[1].value
    count = operands[2].value

    width = get_max_operand_size(operands)
    carry, result = arm_utils.lsr(value, count, width=width)

    if instruction.flag_update:
        if count:
            cpu_context.registers.c = carry
        cpu_context.registers.z = int(result == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.jcccontext.update_flag_opnds(["c", "z", "n"], operands)

    logger.debug("(0x%X >> 0x%X) = 0x%X", value, count, result)
    operands[0].value = result


@opcode
def ROR(cpu_context: ProcessorContext, instruction: Instruction):
    """Rotate right"""
    operands = instruction.operands
    value = operands[1].value
    count = operands[2].value

    width = get_max_operand_size(operands)
    carry, result = arm_utils.ror(value, count, width=width)

    if instruction.flag_update:
        if count:
            cpu_context.registers.c = carry
        cpu_context.registers.z = int(result == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.jcccontext.update_flag_opnds(["c", "z", "n"], operands)

    logger.debug("(0x%X ror 0x%X) = 0x%X", value, count, result)
    operands[0].value = result


@opcode
def EOR(cpu_context: ProcessorContext, instruction: Instruction):
    """XOR operands"""
    operands = instruction.operands
    result = operands[1].value ^ operands[2].value

    if instruction.flag_update:
        width = get_max_operand_size(operands)
        cpu_context.registers.z = int(result == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.jcccontext.update_flag_opnds(["z", "n"], operands)

    operands[0].value = result


#endregion

#region Sign-extend and Zero-extend


@opcode
def SXT(cpu_context: ProcessorContext, instruction: Instruction):
    """Sign-extend"""
    operands = instruction.operands
    operands[0].value = utils.sign_extend(operands[1].value & 0xffff, 2, 4)


@opcode
def UXT(cpu_context: ProcessorContext, instruction: Instruction):
    """Zero-extend"""
    operands = instruction.operands
    operands[0].value = operands[1].value & 0xffff


#endregion

#region Arithmetic (shifted register)


@opcode("neg")
@opcode("ngc")
def NEG(cpu_context: ProcessorContext, instruction: Instruction):
    """Negate (and set flags)"""
    operands = instruction.operands
    value = operands[1].value
    if instruction.root_mnem == "ngc":
        value += int(not cpu_context.registers.c)

    result = -value

    if instruction.flag_update:
        width = operands[1].width
        mask = utils.get_mask(width)
        cpu_context.registers.c = int(result & mask != 0)
        cpu_context.registers.z = int(result & mask == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.registers.v = int(utils.sign_bit(value, width) and not utils.sign_bit(result, width))
        cpu_context.jcccontext.update_flag_opnds(["c", "z", "n", "v"], operands)

    logger.debug("-0x%X -> 0x%X -> %s", value, result, operands[0].text)
    operands[0].value = result

#endregion

#region Flag manipulation instructions


@opcode
def CFINV(cpu_context: ProcessorContext, instruction: Instruction):
    """Invert value of the PSTATE.C bit"""
    cpu_context.registers.c = int(not cpu_context.registers.c)


@opcode
def RMIF(cpu_context: ProcessorContext, instruction: Instruction):
    """Rotate, mask insert flags"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SETF8(cpu_context: ProcessorContext, instruction: Instruction):
    """Evaluation of 8-bit flags"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SETF16(cpu_context: ProcessorContext, instruction: Instruction):
    """Evaluation of 16-bit flags SETF8,"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


#endregion

#region Logical (shifted register)


@opcode
def BIC(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitwise bit clear"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def BICS(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitwise bit clear and set flags"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def EON(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitwise exclusive OR NOT"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def MVN(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitwise NOT"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def ORN(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitwise inclusive OR NOT"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Shift (register)


@opcode
def ASRV(cpu_context: ProcessorContext, instruction: Instruction):
    """Arithmetic shift right variable"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LSLV(cpu_context: ProcessorContext, instruction: Instruction):
    """Logical shift left variable"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def LSRV(cpu_context: ProcessorContext, instruction: Instruction):
    """Logical shift right variable"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def RORV(cpu_context: ProcessorContext, instruction: Instruction):
    """Rotate right variable"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


#endregion

#region Multiply

@opcode
def MADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Multiply-add"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def MSUB(cpu_context: ProcessorContext, instruction: Instruction):
    """Multiply-subtract"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def MNEG(cpu_context: ProcessorContext, instruction: Instruction):
    """Multiply-negate"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def MUL(cpu_context: ProcessorContext, instruction: Instruction):
    """Multiply"""
    operands = instruction.operands
    term_1 = operands[1].value
    term_2 = operands[2].value
    result = term_1 * term_2

    if instruction.flag_update:
        width = get_max_operand_size(operands)
        cpu_context.registers.z = int(result == 0)
        cpu_context.registers.n = utils.sign_bit(result, width)
        cpu_context.jcccontext.update_flag_opnds(["z", "n"], operands)

    logger.debug("0x%X * 0x%X = 0x%X", term_1, term_2, result)
    operands[0].value = result


@opcode
def SMADDL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed multiply-add long"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMSUBL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed multiply-subtract long"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMNEGL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed multiply-negate long"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMULL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed multiply long"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMULH(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed multiply high"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMADDL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned multiply-add long"""
    operands = instruction.operands
    opvalue2 = operands[1].value
    opvalue3 = operands[2].value
    opvalue4 = operands[3].value
    result = (opvalue2 * opvalue3) + opvalue4

    logger.debug("(0x%X * 0x%X) + 0x%X = 0x%X", opvalue2, opvalue2, opvalue4, result)
    operands[0].value = result


@opcode
def UMSUBL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned multiply-subtract long"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMNEGL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned multiply-negate long"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMULL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned multiply long"""
    operands = instruction.operands
    term_1 = operands[-2].value
    term_2 = operands[-1].value
    result = term_1 * term_2

    logger.debug("0x%X * 0x%X = 0x%X", term_1, term_2, result)

    if len(operands) > 3:
        operands[0].value = result & 0xffffffff
        operands[1].value = (result >> 32) & 0xffffffff
    else:
        operands[0].value = result


@opcode
def UMULH(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned multiply high"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Divide


@opcode
def SDIV(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed divide"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UDIV(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned divide"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region CRC32


@opcode
def CRC32(cpu_context: ProcessorContext, instruction: Instruction):
    """CRC-32 sum"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CRC32C(cpu_context: ProcessorContext, instruction: Instruction):
    """CRC-32C sum"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

# region Bit operation


@opcode
def CLS(cpu_context: ProcessorContext, instruction: Instruction):
    """Count leading sign bits"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CLZ(cpu_context: ProcessorContext, instruction: Instruction):
    """Count leading zero bits"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def RBIT(cpu_context: ProcessorContext, instruction: Instruction):
    """Reverse bit order"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def REV(cpu_context: ProcessorContext, instruction: Instruction):
    """Reverse bytes in register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def REV16(cpu_context: ProcessorContext, instruction: Instruction):
    """Reverse bytes in halfwords"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def REV32(cpu_context: ProcessorContext, instruction: Instruction):
    """Reverses bytes in words"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def REV64(cpu_context: ProcessorContext, instruction: Instruction):
    """Reverse bytes in register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Conditional select


@opcode
def CSEL(cpu_context: ProcessorContext, instruction: Instruction):
    """Conditional select"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CSINC(cpu_context: ProcessorContext, instruction: Instruction):
    """Conditional select increment"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CSINV(cpu_context: ProcessorContext, instruction: Instruction):
    """Conditional select inversion"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CSNEG(cpu_context: ProcessorContext, instruction: Instruction):
    """Conditional select negation"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CSET(cpu_context: ProcessorContext, instruction: Instruction):
    """Conditional set"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CSETM(cpu_context: ProcessorContext, instruction: Instruction):
    """Conditional set mask"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CINC(cpu_context: ProcessorContext, instruction: Instruction):
    """Conditional increment"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CINV(cpu_context: ProcessorContext, instruction: Instruction):
    """Conditional invert"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CNEG(cpu_context: ProcessorContext, instruction: Instruction):
    """Conditional negate"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Conditional comparison


@opcode
def CCMN(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Conditional compare negative (register)
    Conditional compare negative (immediate)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CCMP(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Conditional compare (register)
    Conditional compare (immediate)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Floating-point move (register)


@opcode
def FMOV(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Floating-point move register without conversion
    Floating-point move to or from general-purpose register without conversion
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Convert floating-point precision

@opcode
def FCVT(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point convert precision (scalar)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Convert between floating-point and integer or fixed-point


@opcode
def FCVTAS(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar convert to signed integer, rounding to nearest with ties to away (scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTAU(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar convert to unsigned integer, rounding to nearest with ties to away (scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTMS(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar convert to signed integer, rounding toward minus infinity (scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTMU(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar convert to unsigned integer, rounding toward minus infinity (scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTNS(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar convert to signed integer, rounding to nearest with ties to even (scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTNU(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar convert to unsigned integer, rounding to nearest with ties to even (scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTPS(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar convert to signed integer, rounding toward positive infinity (scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTPU(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar convert to unsigned integer, rounding toward positive infinity (scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTZS(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Floating-point scalar convert to signed integer, rounding toward zero (scalar form)
    Floating-point scalar convert to signed fixed-point, rounding toward zero (scalar form)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTZU(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Floating-point scalar convert to unsigned integer, rounding toward zero (scalar form)
    Floating-point scalar convert to unsigned fixed-point, rounding toward zero (scalar form)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FJCVTZS(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point Javascript convert to signed fixed-point, rounding toward zero"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SCVTF(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Signed integer scalar convert to floating-point, using the current rounding mode (scalar form)
    Signed integer fixed-point convert to floating-point, using the current rounding mode (scalar form)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UCVTF(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Unsigned integer scalar convert to floating-point, using the current rounding mode (scalar form)
    Unsigned integer fixed-point convert to floating-point, using the current rounding mode (scalar form)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Floating-point round to integer


@opcode
def FRINTA(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point round to integer, to nearest with ties to away"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FRINTI(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point round to integer, using current rounding mode"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FRINTM(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point round to integer, toward minus infinity"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FRINTN(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point round to integer, to nearest with ties to even"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FRINTP(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point round to integer, toward positive infinity"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FRINTX(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point round to integer exact, using current rounding mode"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FRINTZ(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point round to integer, toward zero"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Floating-point multiply-add


@opcode
def FMADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar fused multiply-add"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMSUB(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar fused multiply-subtract"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FNMADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar negated fused multiply-add"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FNMSUB(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar negated fused multiply-subtract"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Floating-point arithmetic (one source)


@opcode
def FABS(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar absolute value"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FNEG(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar negate"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FSQRT(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar square root"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Floating-point arithmetic (two sources)


@opcode
def FADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar add"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FDIV(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar divide"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMUL(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar multiply"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FNMUL(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar multiply-negate"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FSUB(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar subtract"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Floating-point minimum and maximum


@opcode
def FMAX(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar maximum"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMAXNM(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar maximum number"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMIN(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar minimum"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMINNM(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar minimum number"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Floating-point comparison


@opcode
def FCMP(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point quiet compare"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCMPE(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point signaling compare"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCCMP(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point conditional quiet compare"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCCMPE(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point conditional signaling compare"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region Floating-point conditional select


@opcode
def FCSEL(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point scalar conditional select"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD move


@opcode
def DUP(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Duplicate vector element to vector or scalar
    Duplicate general-purpose register to vector
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def INSa(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Insert vector element from another vector element
    Insert vector element from general-purpose register INS (general) on page C7-16
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMOV(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned move vector element to general-purpose register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMOV(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed move vector element to general-purpose register"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD arithmetic


@opcode
def BIF(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitwise insert if false (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def BIT(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitwise insert if true (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def BSL(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitwise select (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FABD(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point absolute difference (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMLA(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point fused multiply-add (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMLAL(cpu_context: ProcessorContext, instruction: Instruction):
    """FMLAL2 Floating-point fused multiply-add long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMLS(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point fused multiply-subtract (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMLSL(cpu_context: ProcessorContext, instruction: Instruction):
    """FMLSL2 Floating-point fused multiply-subtract long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMULX(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point multiply extended (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FRECPS(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point reciprocal step (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FRSQRTS(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point reciprocal square root step (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def MLA(cpu_context: ProcessorContext, instruction: Instruction):
    """Multiply-add (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def MLS(cpu_context: ProcessorContext, instruction: Instruction):
    """Multiply-subtract (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def PMUL(cpu_context: ProcessorContext, instruction: Instruction):
    """Polynomial multiply (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SABA(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed absolute difference and accumulate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SABD(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed absolute difference (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SHADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed halving add (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SHSUB(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed halving subtract (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMAX(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed maximum (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMIN(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed minimum (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating add (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQDMULH(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating doubling multiply returning high half (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQRSHL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating rounding shift left (register) (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQRDMLAH(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating rounding doubling multiply accumulate returning high half"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQRDMLSH(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating rounding doubling multiply subtract returning high half"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQRDMULH(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating rounding doubling multiply returning high half (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQSHL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating shift left (register) (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQSUB(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating subtract (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SRHADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed rounding halving add (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SRSHL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed rounding shift left (register) (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SSHL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed shift left (register) (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UABA(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned absolute difference and accumulate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UABD(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned absolute difference (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UHADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned halving add (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UHSUB(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned halving subtract (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMAX(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned maximum (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMIN(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned minimum (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UQADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned saturating add (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UQRSHL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned saturating rounding shift left (register) (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UQSHL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned saturating shift left (register) (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UQSUB(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned saturating subtract (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def URHADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned rounding halving add (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def URSHL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned rounding shift left (register) (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def USHL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned shift left (register) (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD compare


@opcode
def CMEQ(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Compare bitwise equal (vector and scalar form)
    Compare bitwise equal to zero (vector and scalar form)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CMHS(cpu_context: ProcessorContext, instruction: Instruction):
    """Compare unsigned higher or same (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CMGE(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Compare signed greater than or equal (vector and scalar form)
    Compare signed greater than or equal to zero (vector and scalar form)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CMHI(cpu_context: ProcessorContext, instruction: Instruction):
    """Compare unsigned higher (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CMGT(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Compare signed greater than (vector and scalar form)
    Compare signed greater than zero (vector and scalar form)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CMLE(cpu_context: ProcessorContext, instruction: Instruction):
    """Compare signed less than or equal to zero (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CMLT(cpu_context: ProcessorContext, instruction: Instruction):
    """Compare signed less than zero (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CMTST(cpu_context: ProcessorContext, instruction: Instruction):
    """Compare bitwise test bits nonzero (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCMEQ(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Floating-point compare equal (vector and scalar form)
    Floating-point compare equal to zero (vector and scalar form)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCMGE(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Floating-point compare greater than or equal (vector and scalar form)
    Floating-point compare greater than or equal to zero (vector and scalar form)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCMGT(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Floating-point compare greater than (vector and scalar form)
    Floating-point compare greater than zero (vector and scalar form)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCMLE(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point compare less than or equal to zero (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCMLT(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point compare less than zero (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FACGE(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point absolute compare greater than or equal (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FACGT(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point absolute compare greater than (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD widening and narrowing arithmetic


@opcode
def ADDHN(cpu_context: ProcessorContext, instruction: Instruction):
    """Add returning high, narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def ADDHN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Add returning high, narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def PMULL(cpu_context: ProcessorContext, instruction: Instruction):
    """Polynomial multiply long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def PMULL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Polynomial multiply long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def RADDHN(cpu_context: ProcessorContext, instruction: Instruction):
    """Rounding add returning high, narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def RADDHN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Rounding add returning high, narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def RSUBHN(cpu_context: ProcessorContext, instruction: Instruction):
    """Rounding subtract returning high, narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def RSUBHN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Rounding subtract returning high, narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SABAL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed absolute difference and accumulate long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SABAL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed absolute difference and accumulate long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SABDL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed absolute difference long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SABDL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed absolute difference long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SADDL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed add long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SADDL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed add long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SADDW(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed add wide (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SADDW2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed add wide (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMLAL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed multiply-add long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMLAL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed multiply-add long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMLSL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed multiply-subtract long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMLSL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed multiply-subtract long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMULL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed multiply long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQDMLAL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating doubling multiply-add long (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQDMLAL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating doubling multiply-add long (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQDMLSL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating doubling multiply-subtract long (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQDMLSL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating doubling multiply-subtract long (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQDMULL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating doubling multiply long (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQDMULL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating doubling multiply long (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SSUBL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed subtract long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SSUBL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed subtract long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SSUBW(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed subtract wide (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SSUBW2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed subtract wide (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SUBHN(cpu_context: ProcessorContext, instruction: Instruction):
    """Subtract returning high, narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SUBHN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Subtract returning high, narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UABAL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned absolute difference and accumulate long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UABAL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned absolute difference and accumulate long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UABDL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned absolute difference long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UABDL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned absolute difference long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UADDL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned add long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UADDL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned add long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UADDW(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned add wide (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UADDW2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned add wide (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMLAL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned multiply-add long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMLAL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned multiply-add long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMLSL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned multiply-subtract long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMLSL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned multiply-subtract long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMULL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned multiply long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def USUBL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned subtract long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def USUBL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned subtract long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def USUBW(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned subtract wide (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def USUBW2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned subtract wide (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD unary arithmetic


@opcode
def ABS(cpu_context: ProcessorContext, instruction: Instruction):
    """Absolute value (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def CNT(cpu_context: ProcessorContext, instruction: Instruction):
    """Population count per byte (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTL(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point convert to higher precision long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point convert to higher precision long (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTN(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point convert to lower precision narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point convert to lower precision narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTXN(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point convert to lower precision narrow, rounding to odd (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCVTXN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point convert to lower precision narrow, rounding to odd (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FRECPE(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point reciprocal estimate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FRECPX(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point reciprocal square root (scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FRSQRTE(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point reciprocal square root estimate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def NOT(cpu_context: ProcessorContext, instruction: Instruction):
    """Bitwise"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SADALP(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed add and accumulate long pairwise (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SADDLP(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed add long pairwise (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQABS(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating absolute value (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQNEG(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating negate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQXTN(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating extract narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQXTN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating extract narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQXTUN(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating extract unsigned narrow (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQXTUN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating extract unsigned narrow (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SUQADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating accumulate of unsigned value (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SXTL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed extend long"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SXTL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed extend long"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UADALP(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned add and accumulate long pairwise (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UADDLP(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned add long pairwise (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UQXTN(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned saturating extract narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UQXTN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned saturating extract narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def URECPE(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned reciprocal estimate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def URSQRTE(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned reciprocal square root estimate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def USQADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned saturating accumulate of signed value (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UXTL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned extend long"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UXTL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned extend long"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def XTN(cpu_context: ProcessorContext, instruction: Instruction):
    """Extract narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def XTN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Extract narrow (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD by element arithmetic


@opcode
def FMLAL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point fused multiply-add long (vector form) FMLAL,"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMLSL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point fused multiply-subtract long (vector form) FMLSL,"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD permute


@opcode
def EXT(cpu_context: ProcessorContext, instruction: Instruction):
    """Extract vector from a pair of vectors"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def TRN1(cpu_context: ProcessorContext, instruction: Instruction):
    """Transpose vectors (primary)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def TRN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Transpose vectors (secondary)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UZP1(cpu_context: ProcessorContext, instruction: Instruction):
    """Unzip vectors (primary)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UZP2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unzip vectors (secondary)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def ZIP1(cpu_context: ProcessorContext, instruction: Instruction):
    """Zip vectors (primary)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def ZIP2(cpu_context: ProcessorContext, instruction: Instruction):
    """Zip vectors (secondary)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD immediate


@opcode
def MOVI(cpu_context: ProcessorContext, instruction: Instruction):
    """Move immediate"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def MVNI(cpu_context: ProcessorContext, instruction: Instruction):
    """Move inverted immediate"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD shift (immediate)


@opcode
def RSHRN(cpu_context: ProcessorContext, instruction: Instruction):
    """Rounding shift right narrow immediate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def RSHRN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Rounding shift right narrow immediate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SHL(cpu_context: ProcessorContext, instruction: Instruction):
    """Shift left immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SHLL(cpu_context: ProcessorContext, instruction: Instruction):
    """Shift left long (by element size) (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SHLL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Shift left long (by element size) (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SHRN(cpu_context: ProcessorContext, instruction: Instruction):
    """Shift right narrow immediate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SHRN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Shift right narrow immediate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SLI(cpu_context: ProcessorContext, instruction: Instruction):
    """Shift left and insert immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQRSHRN(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating rounded shift right narrow immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQRSHRN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating rounded shift right narrow immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQRSHRUN(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating shift right unsigned narrow immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQRSHRUN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating shift right unsigned narrow immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQSHLU(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating shift left unsigned immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQSHRN(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating shift right narrow immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQSHRN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating shift right narrow immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQSHRUN(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating shift right unsigned narrow immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SQSHRUN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed saturating shift right unsigned narrow immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SRI(cpu_context: ProcessorContext, instruction: Instruction):
    """Shift right and insert immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SRSHR(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed rounding shift right immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SRSRA(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed rounding shift right and accumulate immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SSHLL(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed shift left long immediate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SSHLL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed shift left long immediate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SSHR(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed shift right immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SSRA(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed integer shift right and accumulate immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UQRSHRN(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned saturating rounded shift right narrow immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UQRSHRN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned saturating rounded shift right narrow immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UQSHRN(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned saturating shift right narrow immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UQSHRN2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned saturating shift right narrow immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def URSHR(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned rounding shift right immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def URSRA(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned integer rounding shift right and accumulate immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def USHLL(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned shift left long immediate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def USHLL2(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned shift left long immediate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def USHR(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned shift right immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def USRA(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned shift right and accumulate immediate (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD reduce (across vector lanes)


@opcode
def ADDV(cpu_context: ProcessorContext, instruction: Instruction):
    """Add (across vector)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMAXNMV(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point maximum number (across vector)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMAXV(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point maximum (across vector)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMINNMV(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point minimum number (across vector)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMINV(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point minimum (across vector)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SADDLV(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed add long (across vector)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMAXV(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed maximum (across vector)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMINV(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed minimum (across vector)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UADDLV(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned add long (across vector)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMAXV(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned maximum (across vector)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMINV(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned minimum (across vector)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD pairwise arithmetic


@opcode
def ADDP(cpu_context: ProcessorContext, instruction: Instruction):
    """Add pairwise (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FADDP(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point add pairwise (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMAXNMP(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point maximum number pairwise (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMAXP(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point maximum pairwise (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMINNMP(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point minimum number pairwise (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FMINP(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point minimum pairwise (vector and scalar form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMAXP(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed maximum pairwise"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def SMINP(cpu_context: ProcessorContext, instruction: Instruction):
    """Signed minimum pairwise"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMAXP(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned maximum pairwise"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UMINP(cpu_context: ProcessorContext, instruction: Instruction):
    """Unsigned minimum pairwise"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD dot product


@opcode
def SDOT(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Signed dot product (vector form)
    Signed dot product (indexed form)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def UDOT(cpu_context: ProcessorContext, instruction: Instruction):
    """
    Unsigned dot product (vector form)
    Unsigned dot product (indexed form)
    """
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD table lookup


@opcode
def TBL(cpu_context: ProcessorContext, instruction: Instruction):
    """Table vector lookup"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def TBX(cpu_context: ProcessorContext, instruction: Instruction):
    """Table vector lookup extension"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)

#endregion

#region SIMD complex number arithmetic


@opcode
def FCADD(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point complex add"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


@opcode
def FCMLA(cpu_context: ProcessorContext, instruction: Instruction):
    """Floating-point complex multiply accumulate (vector form)"""
    logger.debug("%s instruction not currently implemented.", instruction.mnem)


#endregion

#region Global helper functions


# TODO: Move to Instruction.
def get_max_operand_size(operands):
    """
    Given the list of named tuples containing the operand value and bit width, determine the largest bit width.

    :param operands: list of Operand objects

    :return: largest operand width
    """
    return max(operand.width for operand in operands)

#endregion



"""
Interface for memory management.
"""
from __future__ import annotations

import io
import os
from typing import TYPE_CHECKING, Tuple, List, Optional, Iterable, Union
import warnings

import collections
import logging
from copy import deepcopy

import dragodis

from . import utils
from .constants import *

if TYPE_CHECKING:
    from rugosa.emulation.cpu_context import ProcessorContext

logger = logging.getLogger(__name__)


class Stream(io.RawIOBase):
    """
    Creates a file-like stream of the emulated memory.
    """

    def __init__(self, memory: Memory, start: int):
        self._memory = memory
        self._start = start
        self._offset = 0
        # Figure out which block we are in and use the end of the block as end.
        for base_address, size in memory.blocks:
            if start in range(base_address, base_address + size):
                self._end = size
                break
        else:
            raise RuntimeError(f"Failed to determine end address for memory stream starting at 0x{start:08X}")

    def readable(self) -> bool:
        return True

    def writeable(self) -> bool:
        return True

    def seekable(self) -> bool:
        return True

    def read(self, size: int = -1) -> bytes:
        if size == -1:
            return self.readall()
        size = min(self._end - self._offset, size)
        if size <= 0:
            return b""
        address = self.tell_address()
        data = self._memory.read(address, size)
        self._offset += len(data)
        return data

    def readline(self, size: int = 1) -> bytes:
        address = self.tell_address()
        end = self._memory.find(b"\n", start=address)
        if end == -1:
            return b""
        return self.read(end - address)

    def readall(self) -> bytes:
        return self.read(self._end - self._offset)

    def write(self, data: bytes) -> int:
        address = self.tell_address()
        num_bytes = self._memory.write(address, data)
        self._offset += num_bytes
        return num_bytes

    def tell(self) -> int:
        return self._offset

    def tell_address(self) -> int:
        return self._start + self._offset

    def seek(self, offset: int, whence=os.SEEK_SET) -> int:
        if whence == os.SEEK_SET:
            if offset < 0:
                raise ValueError(f"Offset must be positive.")
            self._offset = offset
        elif whence == os.SEEK_CUR:
            self._offset = max(0, self._offset + offset)
        elif whence == os.SEEK_END:
            self._offset = min(self._end, self._end + offset)
        return self._offset

    def seek_address(self, address: int) -> int:
        return self.seek(address - self._start)


class PageMap(collections.defaultdict):
    """
    Dictionary of page indexes to pages.

    Creates a new page when missing.
    New pages uses the bytes from the IDB if in a segment.
    Segment pages will be mapped, but data retrieval will be delayed until the page
    is requested. (Helps to avoid unnecessary processing of large unused data segments.)
    """

    PAGE_SIZE = 0x1000

    def __init__(self, dis: dragodis.Disassembler, map_segments=True, _cache=None):
        # Setting default_factory to None, because we have overwritten it in __missing__()
        super().__init__(None)

        # Cache of segment pages.
        # Used to prevent multiple calls to pull data from the disassembler.
        if _cache is None:
            _cache = {}
        self._segment_cache = _cache

        self._dis = dis
        if map_segments:
            self.map_segments()

    def __deepcopy__(self, memo):
        copy = PageMap(self._dis, map_segments=False, _cache=self._segment_cache)
        memo[id(self)] = copy
        copy.update({index: (page[:] if page is not None else None) for index, page in self.items()})
        return copy

    def __missing__(self, page_index):
        """
        Creates a new page when index first encountered.

        :return: page
        :rtype: bytearray
        """
        ret = self[page_index] = self._new_page(page_index)
        return ret

    def __getitem__(self, page_index):
        try:
            page = super().__getitem__(page_index)
        except KeyError:
            return self.__missing__(page_index)

        # If page is None, that means this was set for delayed retrieval.
        # Retrieve page now that it is being requested.
        if page is None:
            return self.__missing__(page_index)

        return page

    def _is_delayed(self, page_index):
        """Determines if page is set for delayed retrieval."""
        return page_index in self and super().__getitem__(page_index) is None

    def map_segments(self):
        """Sets segment pages for delayed retrieval"""
        for segment in self._dis.segments:
            if segment.initialized:
                for page_index in range(segment.start >> 12, ((segment.end - 1) >> 12) + 1):
                    self[page_index] = None

    def _new_page(self, page_index: int) -> bytearray:
        """
        Creates a new page based on index.

        :return: page
        """
        if page_index in self._segment_cache:
            return self._segment_cache[page_index][:]

        start = page_index * self.PAGE_SIZE

        # TODO: Keep segment memory windows open and read from them.
        # If page was set for delayed retrieval it is coming from segment data, so pull from disassembler.
        # Update this check if ever use delayed retrieval for non-segment data.
        if self._is_delayed(page_index):
            logger.debug("Reading segment data 0x%X -> 0x%X from disassembler", start, start + self.PAGE_SIZE)
            page = bytearray(self._dis.get_bytes(start, self.PAGE_SIZE, default=0))
            self._segment_cache[page_index] = page[:]  # cache page first
            return page

        # If range is not in a segment, provide a page of all zeros.
        return bytearray(self.PAGE_SIZE)

    def peek(self, page_index: int) -> bytearray:
        """
        Returns the page for the given page index.
        If page doesn't exist, it creates the page but doesn't set it in the map.

        .. warning:: If you are using this you shouldn't try to modify the page since its
            effects may not propagate to the map.
            (ie. this is a read-only copy)

        :param page_index: Index of page to look at.
        :return: page
        """
        if page_index in self and not self._is_delayed(page_index):
            return self[page_index]
        return self._new_page(page_index)


class Memory:
    """
    Class which implements the CPU memory controller backed by the segment data in the input file.

    This class provides a read() and write() function for CPU emulation.
    If a memory address has not been written to, null bytes will be returned.
    """

    PAGE_SIZE = PageMap.PAGE_SIZE

    # Slack space between heap allocations.
    HEAP_SLACK = 0x3000

    # maximum amount of memory allowed to read/write
    # (if we are reading/writing more than ~ 268 MB we have bigger problems.)
    MAX_MEM_READ = 0x10000000
    MAX_MEM_WRITE = 0x10000000

    def __init__(self, cpu_context: ProcessorContext, _cache=None):
        """Initializes Memory object."""
        self._cpu_context = cpu_context
        self._pages = PageMap(cpu_context.emulator.disassembler, _cache=cpu_context.emulator._memory_cache)
        # A map of base addresses to size for heap allocations.
        self._heap_base = cpu_context.emulator.disassembler.max_address
        self._heap_allocations = {}

    def __deepcopy__(self, memo):
        klass = self.__class__
        copy = klass.__new__(klass)
        memo[id(self)] = copy

        copy._cpu_context = deepcopy(self._cpu_context, memo)
        copy._pages = deepcopy(self._pages, memo)
        copy._heap_allocations = self._heap_allocations.copy()
        copy._heap_base = self._heap_base

        return copy

    def open(self, start: int = None) -> Stream:
        """
        Opens memory as a file-like stream.

        :param start: Starting address for the window of data. (defaults to the address of the first allocated block)
        """
        if start is None:
            blocks = self.blocks
            if not blocks:
                raise ValueError("No memory blocks have been allocated.")
            start, _ = blocks[0]
        return Stream(self, start)

    @property
    def blocks(self) -> List[Tuple[int, int]]:
        """
        Returns a list of tuples containing the base address and size for
        contiguous blocks of memory.
        """
        # Collect ranges of continuous memory.
        memory_ranges = []
        base_address = None
        size = 0
        for page_index in sorted(self._pages):
            # First page of block?
            if base_address is None:
                base_address = page_index << 12

            size += self.PAGE_SIZE

            # Found end of continuous block of memory?
            if page_index + 1 not in self._pages:
                memory_ranges.append((base_address, size))
                base_address = None
                size = 0

        # Store last block
        if base_address is not None:
            memory_ranges.append((base_address, size))

        return memory_ranges

    def __str__(self):
        """
        Print information about current memory map.
        """
        # Create text output.
        _just = 25 if self._cpu_context.bitness == 32 else 50
        _hex_fmt = "0x{:08X}" if self._cpu_context.bitness == 32 else "0x{:016X}"
        title = f"{'Base Address'.ljust(_just)}{'Address Range'.ljust(_just)}{'Size'}"
        memory_ranges = []
        for base_address, size in self.blocks:
            memory_ranges.append(
                "{}{}{}".format(
                    _hex_fmt.format(base_address).ljust(_just),
                    "{} - {}".format(_hex_fmt.format(base_address), _hex_fmt.format(base_address + size)).ljust(_just),
                    size,
                )
            )

        return "{}\n{}\n".format(title, "\n".join(memory_ranges))

    def is_mapped(self, address: int) -> bool:
        return address >> 12 in self._pages

    def alloc(self, size: int) -> int:
        """
        Allocates heap region with size number of bytes.

        :param size: Number of bytes to allocate.
        :return: starting address of allocated memory.
        """
        # Allocate from HEAP_BASE if this is our first allocation.
        if not self._heap_allocations:
            address = self._heap_base
        # Otherwise, use the largest base address not used.
        # TODO: We may want to reuse previously freed space in the future.
        else:
            max_base_address = max(self._heap_allocations)
            heap_size = self._heap_allocations[max_base_address]
            address = max_base_address + heap_size + self.HEAP_SLACK

        # NOTE: We are just going to record that the memory as been allocated
        # but not actually trigger any data from being written. (The calls to write() will do that)
        # This helps to prevent us from wasting (real) memory if someone allocates
        # a huge amount of memory but only uses a small amount.
        self._heap_allocations[address] = size
        logger.debug("Allocated %d bytes at 0x%08X", size, address)
        return address

    def realloc(self, address: int, size: int) -> int:
        """
        Reallocates heap region with size number of bytes.

        :param address: base address to reallocate.
        :param size: Number of bytes to allocate.
        :return: address of the reallocated memory block.
        :raises ValueError: If given address is not allocated.
        """
        # Passed in address should be the base address of a previously allocated memory region.
        if address not in self._heap_allocations:
            raise ValueError(f"0x{address:X} address is not allocated.")

        previous_size = self._heap_allocations[address]

        # See if we need to relocate the heap address.
        if size > previous_size:
            for base_address in sorted(self._heap_allocations):
                if address < base_address < address + size:
                    # We need to relocate the memory block.
                    new_address = self.alloc(size)

                    # Copy over data from previous allocation.
                    # Since relocation is very rare, we will accept the loss in cycles
                    # if we end up writing empty data.
                    self.write(new_address, self.read(address, previous_size))

                    # Don't free the old, because the user may want to search it.
                    logger.debug("Relocated 0x%08X -> 0x%08X", address, new_address)

                    # Record a memory copy since the pointer has changed.
                    self._cpu_context.memory_copies[self._cpu_context.ip].append((address, new_address, size))

                    return new_address

        # Otherwise we just need to adjust the size.
        if previous_size != size:
            logger.debug(
                "Reallocating heap size at 0x%08X from %d to %d bytes.",
                address, previous_size, size
            )
            self._heap_allocations[address] = size
        return address

    def read(self, address: int, size: int) -> bytes:
        """
        Reads data from given address.

        :param address: Address to read data from.
        :param size: Number of bytes to read.

        :return: byte string of read data.
        :raises ValueError: If address or size is negative.
        """
        if address < 0:
            raise ValueError(f"Address must be a positive integer. Got 0x{address:08X}")
        if size < 0:
            raise ValueError("Size must be a positive integer.")
        if size > self.MAX_MEM_READ:
            logger.error(
                "Attempted to read %d bytes from 0x%08X. "
                "Ignoring request and reading the first %d bytes instead.",
                size, address, self.MAX_MEM_READ
            )
            size = self.MAX_MEM_READ

        logger.debug("Reading %d bytes from 0x%08X", size, address)

        page_index = address >> 12
        page_offset = address & 0xFFF

        # Read data from pages.
        out = bytearray()
        while size:
            # We don't want to trigger a page creation on read().. only write().
            page = self._pages.peek(page_index)
            read_bytes = page[page_offset : page_offset + size]
            out += read_bytes
            size -= len(read_bytes)
            page_offset = 0
            page_index += 1
        out = bytes(out)

        # logger.debug("Read: %r", out[:100])  # Enable this when debugging.
        return out

    def write(self, address: int, data: bytes) -> int:
        """
        Writes data to given address.

        :param address: Address to write data to.
        :param data: data to write
        :returns: Number of bytes written.

        :raises ValueError: If given address is negative.
        :raises TypeError: If given data type is invalid.
        """
        if address < 0:
            raise ValueError(f"Address must be a positive integer. Got 0x{address:08X}")

        size = len(data)
        if size > self.MAX_MEM_WRITE:
            logger.error(
                "Attempted to write %d bytes from 0x%08X. "
                "Ignoring request and writing the first %d bytes instead.",
                size, address, self.MAX_MEM_WRITE
            )
            data = data[: self.MAX_MEM_WRITE]
        size = len(data)

        logger.debug("Writing %d bytes to 0x%08X", size, address)
        # logger.debug("Writing: %r", data[:100])   # Enable this when debugging.

        page_index = address >> 12
        page_offset = address & 0xFFF

        # Write data into pages.
        while data:
            page = self._pages[page_index]
            split_index = self.PAGE_SIZE - page_offset
            to_write = data[:split_index]
            try:
                page[page_offset : page_offset + len(to_write)] = to_write
            except TypeError:
                raise TypeError("to_write: {} {}".format(type(to_write), repr(to_write)))
            data = data[split_index:]
            page_offset = 0
            page_index += 1

        return size

    def read_data(self, addr: int, size: int = None, data_type=None) -> Union[bytes, int]:
        """
        Reads memory at the specified address, of the specified size and convert
        the resulting data into the specified type.

        :param addr: address to read data from
        :param size: size of data to read
        :param data_type: type of data to be extracted
            (default to BYTE_STRING is size provided or STRING if not)
        """
        if not data_type:
            data_type = STRING if size is None else BYTE_STRING
        if size is None:
            size = 0

        if data_type == STRING:
            null_offset = self.find(b"\0", start=addr)
            # It should always eventually find a null since unmapped pages
            # are all null. If we get -1 we have a bug.
            assert null_offset != -1, "Unable to find a null character!"
            return self.read(addr, null_offset - addr)

        elif data_type == WIDE_STRING:
            # Step by 2 bytes to find 2 nulls on an even alignment.
            # (This helps prevent the need to take endianness into account.)
            null_offset = addr
            while self.read(null_offset, 2) != b"\0\0":
                null_offset += 2

            return self.read(addr, null_offset - addr)

        elif data_type == BYTE_STRING:
            return self.read(addr, size)

        elif data_type == BYTE:
            data = self.read(addr, 1)
            return int.from_bytes(data, self._cpu_context.byteorder)

        elif data_type == WORD:
            data = self.read(addr, 1)
            return int.from_bytes(data, self._cpu_context.byteorder)

        elif data_type == DWORD:
            data = self.read(addr, 4)
            return int.from_bytes(data, self._cpu_context.byteorder)

        elif data_type == QWORD:
            data = self.read(addr, 8)
            return int.from_bytes(data, self._cpu_context.byteorder)

        raise ValueError("Invalid data_type: {!r}".format(data_type))

    def read_string(self, addr: int, wide=False, encoding=None) -> str:
        """
        Helper function for reading and decoding a null terminated string
        at the given address.

        :param addr: address to read data from
        :param wide: Determines if data to be read is a wide or non-wide string.
            If true, data_type is assumed to be a STRING.
            If false, data_type is assumed to be a WIDE_STRING.
        :param encoding: Encoding to use for decoding.
            (defaults to utf8 for non-wide strings, and utf-16-le for wide strings.)
        :return: Decoded string
        """
        if not encoding:
            encoding = "utf-16-le" if wide else "utf8"
        data = self.read_data(addr, data_type=WIDE_STRING if wide else STRING)
        return data.decode(encoding)

    def write_data(self, addr: int, value: Union[str, bytes, int], data_type=None):
        """
        Writes memory at the specified address after converting the value
        into data based on the specified data type.

        :param addr: address to write data to
        :param value: integer or byte string to write
        :param data_type: type of data to convert value from.
            (defaults to BYTE_STRING, STRING, or DWORD based on input data)
        """
        if not data_type:
            if isinstance(value, str):
                data_type = STRING
            elif isinstance(value, bytes):
                data_type = BYTE_STRING
            elif isinstance(value, int):
                data_type = DWORD
            else:
                raise ValueError(f"Invalid data type: {type(value)}")

        if data_type == BYTE_STRING:
            data = value

        elif data_type == STRING:
            data = value
            if isinstance(data, str):
                data = data.encode("utf8")
            data += b"\0"

        elif data_type == WIDE_STRING:
            data = value
            if isinstance(data, str):
                data = data.encode("utf-16-le")
            data += b"\0\0"

        elif data_type == BYTE:
            data = bytes([value])

        elif data_type == WORD:
            value = utils.unsigned(value, 16)
            data = value.to_bytes(2, self._cpu_context.byteorder)

        elif data_type == DWORD:
            value = utils.unsigned(value, 32)
            data = value.to_bytes(4, self._cpu_context.byteorder)

        elif data_type == QWORD:
            value = utils.unsigned(value, 64)
            data = value.to_bytes(8, self._cpu_context.byteorder)

        else:
            raise ValueError(f"Invalid data_type: {repr(data_type)}")

        self.write(addr, data)

    def write_string(self, addr: int, value: str, wide=False, encoding=None):
        """
        Helper function for encoding and writing a null terminated string
        at the given address.

        NOTE: This is basically the reverse of read_string()

        :param addr: address to write data to
        :param value: string to write
        :param wide: Determines if data to be written as a wide or non-wide string.
            If true, data_type is assumed to be a STRING.
            If false, data_type is assumed to be a WIDE_STRING.
        :param encoding: Encoding to use for encoding.
            (defaults to utf8 for non-wide strings, and utf-16-le for wide strings.)
        """
        if not encoding:
            encoding = "utf-16-le" if wide else "utf8"
        value = value.encode(encoding)
        self.write_data(addr, value, data_type=WIDE_STRING if wide else STRING)

    def copy(self, src: int, dst: int, size: int):
        """
        Copy data from src address to dst address.
        (Use this over read()/write() combo in order to allow the context to keep
        track of memory pointer history.)

        :param src: Source address
        :param dst: Destination address
        :param size: Number of bytes to copy over.
        """
        self._cpu_context.memory_copies[self._cpu_context.ip].append((src, dst, size))
        self.write(dst, self.read(src, size))

    def find(self, value: bytes, start: int = 0, end: Optional[int] = None) -> int:
        """
        Searches memory for given value.

        :param value: byte string to search for
        :param start: Starting address to start search (defaults to 0)
        :param end: Optional ending address to end search

        :return int: address where value was located or -1 if not found

        :raises ValueError: If search value is larger than a page.
        """
        # We are not going to handle things that could expand beyond multiple pages
        if len(value) >= self.PAGE_SIZE:
            raise ValueError(f"Search value must be less than {self.PAGE_SIZE} bytes, got {len(value)}")

        if end and end <= start:
            raise ValueError("Ending address must be greater than starting address.")

        page_index = start >> 12
        page_offset = start & 0xFFF

        if end:
            end_page_index = end >> 12
            end_page_offset = end & 0xFFF
        else:
            end_page_index = end_page_offset = None

        page = self._pages.peek(page_index)

        # First search for the entire value in the page.
        if end and end_page_index == page_index:
            # Since we end in this page, don't attempt to read overlap.
            offset = page.find(value, page_offset, end_page_offset)
            if offset <= -1:
                return -1
            return page_index << 12 | offset

        offset = page.find(value, page_offset)
        if offset > -1:
            return page_index << 12 | offset

        # If we can't find it, try again with part of the next page attached
        # to account for data overlapping onto another page.
        _start = max(page_offset, self.PAGE_SIZE - len(value) + 1)
        _end = self.PAGE_SIZE + len(value) - 1
        if end and end_page_index == page_index + 1:
            _end = min(_end, self.PAGE_SIZE + end_page_offset)
            if _end == self.PAGE_SIZE + end_page_offset and _end <= _start:
                return -1

        next_page = self._pages.peek(page_index + 1)
        offset = (page + next_page).find(value, _start, _end)
        if offset > -1:
            return page_index << 12 | offset
        else:
            # Jump to the next mapped page to continue the search.
            # return -1 if there are no more pages beyond.
            for _page_index in sorted(self._pages):
                if _page_index > page_index:
                    # Stop searching pages if we surpass the end.
                    if end and end <= _page_index << 12:
                        return -1
                    return self.find(value, start=_page_index << 12, end=end)
            return -1

    def finditer(self, value: bytes, start: int = 0, end: Optional[int] = None) -> Iterable[int]:
        """
        Searches for all instances of value within memory.
        """
        while True:
            offset = self.find(value, start=start, end=end)
            if offset == -1:
                return
            yield offset
            start = offset + len(value)

    def find_in_segment(self, value: bytes, addr_or_name: Union[int, str]) -> int:
        """
        Searches memory for given value within the range of a specific segment.

        :param bytes value: byte string to search for
        :param addr_or_name: segment name or address within segment.

        :return: address where value was located or -1 if not found
        """
        segment = self._cpu_context.emulator.disassembler.get_segment(addr_or_name)
        return self.find(value, start=segment.start, end=segment.end)

    def finditer_in_segment(self, value: bytes, addr_or_name: Union[int, str]) -> Iterable[int]:
        """
        Searches memory for given value within the range of a specific segment.

        :param bytes value: byte string to search for
        :param addr_or_name: segment name or address withing segment.

        :yields: address where value was located or -1 if not found
        """
        segment = self._cpu_context.emulator.disassembler.get_segment(addr_or_name)
        yield from self.finditer(value, start=segment.start, end=segment.end)

    def find_in_heap(self, value: bytes) -> int:
        """
        Searches memory for given value within the allocated heap.

        :param value: byte string to search for

        :return: address where value was located or -1 if not found
        """
        return self.find(value, start=self._heap_base)

    def finditer_in_heap(self, value: bytes) -> Iterable[int]:
        """
        Searches memory for given value within the allocated heap.

        :param value: byte string to search for

        :yields: address where value was located or -1 if not found
        """
        yield from self.finditer(value, start=self._heap_base)




from rugosa import yara


rule_text = """
rule MyStrings 
{
    strings:
        $string_1 = "Idmmn!Vnsme"
        $string_2 = "Dfla%gpwkv%mji`v%lk%rjji%fijqm+"
    condition:
        any of them
}
"""


def test_match(disassembler):
    rule = yara.compile(source=rule_text)
    matches = rule.match(disassembler)
    assert matches
    assert [match.strings for match in matches] == [
        [(0x40C000, '$string_1', b'Idmmn!Vnsme'),
         (0x40C080, '$string_2', b'Dfla%gpwkv%mji`v%lk%rjji%fijqm+')]
    ]
    matches = rule.match(disassembler, segment=".text")
    assert not matches
    matches = rule.match(disassembler, segment=".data")
    assert matches
    assert [match.strings for match in matches] == [
        [(0x40C000, '$string_1', b'Idmmn!Vnsme'),
         (0x40C080, '$string_2', b'Dfla%gpwkv%mji`v%lk%rjji%fijqm+')]
    ]



from __future__ import annotations
from typing import TYPE_CHECKING

from PyQt5 import QtWidgets

if TYPE_CHECKING:
    from rugosa.ida_plugin.components.emulator import EmulatorForm


class CallHistory:

    def __init__(self, ui: EmulatorForm):
        self._ctx = None
        self.ui = ui
        self.table = ui.call_history_table
        self.table.horizontalHeader().setSectionResizeMode(QtWidgets.QHeaderView.ResizeToContents)
        self.table.itemDoubleClicked.connect(ui.redirect)

    def populate(self, ctx):
        """
        Populate the call history table with current context.
        """
        self._ctx = ctx
        call_history = ctx.call_history
        table = self.table
        table.setRowCount(len(call_history))
        for index, (call_address, func_name, args) in enumerate(call_history):
            args = ", ".join((f"{name}={hex(value)}" if name else hex(value)) for name, value in args)
            func_call = f"{func_name}({args})"
            table.setItem(index, 0, QtWidgets.QTableWidgetItem(hex(call_address)))
            table.setItem(index, 1, QtWidgets.QTableWidgetItem(func_call))



from __future__ import annotations
import logging
from typing import TYPE_CHECKING

import jpype

import dragodis

RugosaChangeListener = jpype.JClass("dc3.rugosa.plugin.RugosaChangeListener")

from rugosa.ghidra_plugin.gui.error_msg import catch_errors
from rugosa.ghidra_plugin.gui.table import TableModel
from rugosa.ghidra_plugin.gui.context_menu import TableContextMenuEntry


if TYPE_CHECKING:
    from javax.swing.event import ChangeEvent
    from rugosa.ghidra_plugin.components.emulator import EmulatorForm


logger = logging.getLogger(__name__)


class ContextMenuEntry(TableContextMenuEntry):

    def __init__(self, function_arguments: FunctionArguments):
        super().__init__(function_arguments.table)
        self._function_arguments = function_arguments

    @property
    def address(self):
        arg = self._function_arguments._arguments[self._component.getSelectedRow()]
        return arg.value


class JumpToDisassembly(ContextMenuEntry):
    NAME = "Jump To Disassembly"

    def enabled(self, row: int, col: int) -> bool:
        ctx = self._function_arguments._ctx
        address = self.address
        return bool(ctx and address)

    @catch_errors
    def clicked(self, row: int, col: int):
        self._function_arguments.form.plugin.goTo(self._function_arguments.form.dis._to_addr(self.address))


class ShowInMemory(ContextMenuEntry):
    NAME = "Show in Memory"

    def enabled(self, row: int, col: int) -> bool:
        ctx = self._function_arguments._ctx
        address = self.address
        return bool(ctx and address and ctx.memory.is_mapped(address))

    @catch_errors
    def clicked(self, row: int, col: int):
        self._function_arguments.form.memory.show_with_data(self.address, 256)


class FunctionArguments:

    MAX_VALUE_SIZE = 256
    HEADERS = ["Ordinal", "Location", "Data Type", "Size", "Name", "Address", "Value", "Referenced Data"]

    def __init__(self, form: EmulatorForm):
        self._ctx = None
        self._arguments = None
        self.form = form
        self.ui = form.ui
        self.table = self.ui.functionArgumentsTable
        self.signature = self.ui.functionSignature
        self.num_args = self.ui.numArgs
        self.num_args.addChangeListener(RugosaChangeListener @ self.num_args_updated)
        self.form.add_context_menu_entry(JumpToDisassembly(self))
        self.form.add_context_menu_entry(ShowInMemory(self))

    @catch_errors
    def num_args_updated(self, event: ChangeEvent):
        if self._ctx:
            # Need to be in transaction because we edit the function parameters in the databaes.
            with self.form.transaction():
                self.populate(self._ctx, num_args=self.num_args.getValue())

    def _populate_table(self, arguments):
        """
        Populates the function arguments table with given arguments.
        """
        self._arguments = arguments
        rows = []
        for arg in arguments:
            row = [""] * len(self.HEADERS)
            row[0] = str(arg.ordinal)
            row[1] = str(arg.location)
            row[2] = arg.type
            row[3] = str(arg.width)
            row[4] = arg.name
            addr = arg.addr
            if addr is not None:
                row[5] = hex(addr)
            row[6] = hex(arg.value)

            # If argument value is a pointer, show the first few bytes of the referenced data
            if self._ctx.memory.is_mapped(arg.value):
                data = self._ctx.memory.read(arg.value, 24)
                row[7] = str(data) + "..."

            rows.append(row)
        self.table.setModel(TableModel(self.HEADERS, rows))

    def populate(self, ctx, num_args=None):
        self._ctx = ctx

        try:
            func_sig = ctx.get_function_signature(num_args=num_args)
            if not func_sig:
                raise dragodis.NotExistError("No Operand.")
            self.num_args.setValue(jpype.JInt(len(func_sig.arguments)))
            self.signature.setText(
                f"{hex(func_sig.address)}: {func_sig.declaration}"
            )
            self._populate_table(func_sig.arguments)

        except dragodis.NotExistError:
            self.signature.setText("Function call not detected")
            self.table.setModel(TableModel(self.HEADERS, []))  # this clears the table
            return

        except Exception as e:
            self.signature.setText(f"Error: {e}")
            self.table.setModel(TableModel(self.HEADERS, []))
            logger.exception(e)
            return





"""
This file contains utility functions utilized throughout the emulation package.
"""
from __future__ import annotations

import logging
import re
import struct

from dragodis import Disassembler, NotExistError
from .exceptions import EmulationError


logger = logging.getLogger(__name__)


def signed(n: int, size: int) -> int:
    """
    Convert an unsigned integer to a signed integer

    :param uint n: value to convert
    :param int size: byte width of n

    :return int: signed conversion
    """
    size *= 8
    if n >> (size - 1):  # Is the hi-bit set?
        return n - (1 << size)
    return n


def unsigned(n: int, size: int) -> int:
    """
    Convert a signed integer to an unsigned integer

    :param sint n: value to convert
    :param in size: byte width of n
        Defaults to architecture addressing size.

    :return int: unsigned conversion
    """
    size *= 8
    return n & ((1 << size) - 1)


def sign_bit(value: int, size: int) -> int:
    """Returns the highest bit with given value and byte width."""
    return (value >> ((8 * size) - 1)) & 0x1


def sign_extend(value: int, orig_size: int, dest_size: int) -> int:
    """
    Calculates the sign extension for a provided value and a specified destination size.

    :param value: value to be sign extended
    :param orig_size: byte width of value
    :param dest_size: byte width of value to extend to.
    :return: value, sign extended
    """
    if dest_size < orig_size:
        raise ValueError(f"Destination size must be larger than original size.")
    orig_size *= 8
    dest_size *= 8

    # Calculate the max value for orig and dest
    orig_mask = (1 << orig_size) - 1
    dest_mask = (1 << dest_size) - 1
    # Create the bit mask
    masknumber = value & orig_mask
    msb = masknumber >> (orig_size - 1)
    # Perform the sign extension
    if msb:
        signextended = ((dest_mask << orig_size) | masknumber) & dest_mask
    else:
        signextended = value & dest_mask

    return signextended


def float_to_int(value: float, precision: int = 2) -> int:
    """
    Given a float value, convert it to its integer hexadecimal equivalent.

    >>> float_to_int(1.0)
    >>> 4607182418800017408

    :param value: float to convert to int equivalent
    :param precision: single or double precision (1 for single, 2 for double)
    :return: int
    :raises: ValueError
    """
    if precision == 1:
        return struct.unpack("H", struct.pack("f", value))[0]
    elif precision == 2:
        return struct.unpack("Q", struct.pack("d", value))[0]
    else:
        raise EmulationError(f"Precision {precision} is not valid.")


def int_to_float(value: int, precision: int = 2) -> float:
    """
    Given an integer value, convert it to its float hexadecimal equivalent.

    >>> int_to_float(4607182418800017408, 8)
    >>> 1.0

    :param int value: integer value to convert to float equivalent
    :param int precision: single or double precision
    :return: int or None
    :raises: ValueError
    """
    if precision == 1:
        return struct.unpack("f", struct.pack("H", value))[0]
    elif precision == 2:
        return struct.unpack("d", struct.pack("Q", value))[0]
    else:
        raise EmulationError(f"Precision {precision} is not valid.")


def get_mask(size: int) -> int:
    """
    Get bit mask based on byte size.

    :param size: number of bytes to obtain mask for

    :return: mask of width size
    """
    return (1 << (8 * size)) - 1


def sanitize_func_name(func_name: str) -> str:
    """Sanitizes the IDA function names to it's core name."""
    # remove the extra "_" IDA likes to add to the function name.
    if func_name.startswith("_"):
        func_name = func_name[1:]

    # Remove the numbered suffix IDA likes to add to duplicate function names.
    func_name = re.sub("_[0-9]+$", "", func_name)

    return func_name


def is_func_ptr(dis: Disassembler, address: int) -> bool:
    """Returns true if the given offset is a function pointer."""
    # As a first check, simply see if the offset is the start of a function.
    try:
        func = dis.get_function(address)
        if func.start == address:
            return True
    except NotExistError:
        pass

    # Next determine if we can successfully get a function signature.
    # TODO: Determine if there is a way to do this without calling get_function_signature()
    #   (Optimization)
    try:
        func_sig = dis.get_function_signature(address)
        return True
    except NotExistError:
        pass

    return False
